from langchain.schema import BaseMessage

from llm.direction import TokenDirection
from llm.model_registry import ModelRegistry
from llm.usage import TokenUsage


class TokenCounter:
    """Счетчик токенов и стоимости."""

    def __init__(
        self,
        model_name: str,
        usage: TokenUsage,
        model_registry: ModelRegistry,
    ) -> None:
        self.model_name = model_name
        self.usage = usage
        self.model_registry = model_registry

    async def count_input_tokens_from_text(self, messages: list[BaseMessage]) -> None:
        """Считает количество токенов из текста, которое отправляются в LLM.

        Args:
            messages (list[BaseMessage]): Сообщения
        """
        tokens = await self.model_registry.get_tokens(self.model_name, messages)
        await self.count_input_tokens(tokens)

    async def count_output_tokens_from_text(self, messages: list[BaseMessage]) -> None:
        """Считает количество токенов из текста, которое получаются из LLM.

        Args:
            messages (list[BaseMessage]): Сообщения
        """
        tokens = await self.model_registry.get_tokens(self.model_name, messages)
        await self.count_output_tokens(tokens)

    async def count_input_tokens(self, tokens: int) -> None:
        """Считает количество токенов, которое отправляются в LLM.

        Args:
            tokens (int): токены
        """
        await self._enc_spendings(tokens)
        self.usage.all_input_tokens += tokens
        self.usage.last_input_tokens = tokens

    async def count_output_tokens(self, tokens: int) -> None:
        """Считает количество токенов, которое получаются из LLM.

        Args:
            tokens (int): токены
        """
        await self._dec_spendings(tokens)
        self.usage.all_output_tokens += tokens
        self.usage.last_output_tokens = tokens

    async def _enc_spendings(self, tokens: int) -> None:
        """Считает расходы в USD при отправке в LLM.

        Args:
            tokens (int): токены
        """
        toc_price = self.model_registry.get_price(
            self.model_name, TokenDirection.ENCODE
        )

        self.usage.all_input_spendings += tokens * toc_price
        self.usage.last_input_spendings = tokens * toc_price

    async def _dec_spendings(self, tokens: int) -> None:
        """Считает расходы в USD при получении из LLM.

        Args:
            tokens (int): токены
        """
        toc_price = self.model_registry.get_price(
            self.model_name, TokenDirection.DECODE
        )

        self.usage.all_output_spendings += tokens * toc_price
        self.usage.last_output_spendings = tokens * toc_price



import warnings
from dataclasses import dataclass
from typing import Any, Callable, Type

from langchain.schema import BaseMessage
from langchain_anthropic import ChatAnthropic
from langchain_cerebras import ChatCerebras
from langchain_deepseek import ChatDeepSeek
from langchain_gigachat import GigaChat
from langchain_google_genai import ChatGoogleGenerativeAI
from langchain_openai import ChatOpenAI
from langchain_xai import ChatXAI

from llm.direction import TokenDirection
from llm.moderations import ModerationPrompt
from llm.token_counters import TokenCounterFactory

LLMClientInstance = (
    ChatOpenAI
    | GigaChat
    | ChatAnthropic
    | ChatGoogleGenerativeAI
    | ChatXAI
    | ChatDeepSeek
    | ChatCerebras
)
LLMClientClass = (
    Type[ChatOpenAI]
    | Type[GigaChat]
    | Type[ChatAnthropic]
    | Type[ChatGoogleGenerativeAI]
    | Type[ChatXAI]
    | Type[ChatDeepSeek]
    | Type[ChatCerebras]
)


@dataclass
class ModelConfig:
    """Конфигурация для конкретной модели"""

    client_class: LLMClientClass
    token_counter: Callable
    pricing: dict[TokenDirection, float]
    moderation: Callable | None = None


class ModelRegistry:
    """Реестр моделей

    - Цены OpenAI: https://platform.openai.com/docs/pricing
    - Цены Gigachat: https://developers.sber.ru/docs/ru/gigachat/tariffs/legal-tariffs
    - Цены Anthropic: https://docs.anthropic.com/en/docs/about-claude/pricing
        - имена моделей https://docs.anthropic.com/en/docs/about-claude/models/overview#model-names
    - Цены Google: https://ai.google.dev/gemini-api/docs/pricing#gemini-2.5-pro-preview
    - Цены XAI: https://docs.x.ai/docs/models
    - Цены DeepSeek: https://api-docs.deepseek.com/quick_start/pricing
    """  # noqa: E501

    def __init__(self, usd_rate: float) -> None:
        self.usd_rate = usd_rate
        self.client: LLMClientInstance = None
        self._models = self._init_models()

    def _init_models(self) -> dict[str, ModelConfig]:
        return {
            # OpenAI
            'gpt-5': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 1.25 / 1_000_000,
                    TokenDirection.DECODE: 10.00 / 1_000_000,
                },
            ),
            'gpt-5-mini': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 0.25 / 1_000_000,
                    TokenDirection.DECODE: 2.00 / 1_000_000,
                },
            ),
            'gpt-5-nano': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 0.05 / 1_000_000,
                    TokenDirection.DECODE: 0.40 / 1_000_000,
                },
            ),
            'gpt-5-chat-latest': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 1.25 / 1_000_000,
                    TokenDirection.DECODE: 10.00 / 1_000_000,
                },
            ),
            'gpt-4.1': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 2.00 / 1_000_000,
                    TokenDirection.DECODE: 8.00 / 1_000_000,
                },
            ),
            'gpt-4.1-mini': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 0.40 / 1_000_000,
                    TokenDirection.DECODE: 1.60 / 1_000_000,
                },
            ),
            'gpt-4.1-nano': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 0.10 / 1_000_000,
                    TokenDirection.DECODE: 0.40 / 1_000_000,
                },
            ),
            'gpt-4.5-preview': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 75.00 / 1_000_000,
                    TokenDirection.DECODE: 150.00 / 1_000_000,
                },
            ),
            'gpt-4o-mini': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 0.15 / 1_000_000,
                    TokenDirection.DECODE: 0.60 / 1_000_000,
                },
            ),
            'gpt-4o': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 2.50 / 1_000_000,
                    TokenDirection.DECODE: 10.00 / 1_000_000,
                },
            ),
            'o3-2025-04-16': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 2.00 / 1_000_000,
                    TokenDirection.DECODE: 8.00 / 1_000_000,
                },
            ),
            'o4-mini-2025-04-16': ModelConfig(
                client_class=ChatOpenAI,
                token_counter=TokenCounterFactory().create_openai_counter(),
                moderation=ModerationPrompt().openai,
                pricing={
                    TokenDirection.ENCODE: 1.10 / 1_000_000,
                    TokenDirection.DECODE: 4.40 / 1_000_000,
                },
            ),
            # Gigachat
            'GigaChat': ModelConfig(
                client_class=GigaChat,
                token_counter=TokenCounterFactory().create_gigachat_counter(),
                pricing={
                    # 5_000 рублей / 25_000_000 токенов / курс доллара
                    TokenDirection.ENCODE: 5_000 / 25_000_000 / self.usd_rate,
                    TokenDirection.DECODE: 5_000 / 25_000_000 / self.usd_rate,
                },
            ),
            'GigaChat-2': ModelConfig(
                client_class=GigaChat,
                token_counter=TokenCounterFactory().create_gigachat_counter(),
                pricing={
                    # 5_000 рублей / 25_000_000 токенов / курс доллара
                    TokenDirection.ENCODE: 5_000 / 25_000_000 / self.usd_rate,
                    TokenDirection.DECODE: 5_000 / 25_000_000 / self.usd_rate,
                },
            ),
            'GigaChat-Pro': ModelConfig(
                client_class=GigaChat,
                token_counter=TokenCounterFactory().create_gigachat_counter(),
                pricing={
                    # 10_500 рублей / 7_000_000 токенов / курс доллара
                    TokenDirection.ENCODE: 10_500 / 7_000_000 / self.usd_rate,
                    TokenDirection.DECODE: 10_500 / 7_000_000 / self.usd_rate,
                },
            ),
            'GigaChat-2-Pro': ModelConfig(
                client_class=GigaChat,
                token_counter=TokenCounterFactory().create_gigachat_counter(),
                pricing={
                    # 10_500 рублей / 7_000_000 токенов / курс доллара
                    TokenDirection.ENCODE: 10_500 / 7_000_000 / self.usd_rate,
                    TokenDirection.DECODE: 10_500 / 7_000_000 / self.usd_rate,
                },
            ),
            'GigaChat-Max': ModelConfig(
                client_class=GigaChat,
                token_counter=TokenCounterFactory().create_gigachat_counter(),
                pricing={
                    # 15_600 рублей / 8_000_000 токенов / курс доллара
                    TokenDirection.ENCODE: 15_600 / 8_000_000 / self.usd_rate,
                    TokenDirection.DECODE: 15_600 / 8_000_000 / self.usd_rate,
                },
            ),
            'GigaChat-2-Max': ModelConfig(
                client_class=GigaChat,
                token_counter=TokenCounterFactory().create_gigachat_counter(),
                pricing={
                    # 15_600 рублей / 8_000_000 токенов / курс доллара
                    TokenDirection.ENCODE: 15_600 / 8_000_000 / self.usd_rate,
                    TokenDirection.DECODE: 15_600 / 8_000_000 / self.usd_rate,
                },
            ),
            # Anthropic
            'claude-3-5-haiku-latest': ModelConfig(
                client_class=ChatAnthropic,
                token_counter=TokenCounterFactory().create_anthropic_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.80 / 1_000_000,
                    TokenDirection.DECODE: 4.00 / 1_000_000,
                },
            ),
            'claude-3-7-sonnet-latest': ModelConfig(
                client_class=ChatAnthropic,
                token_counter=TokenCounterFactory().create_anthropic_counter(),
                pricing={
                    TokenDirection.ENCODE: 3.00 / 1_000_000,
                    TokenDirection.DECODE: 15.00 / 1_000_000,
                },
            ),
            'claude-opus-4-0': ModelConfig(
                client_class=ChatAnthropic,
                token_counter=TokenCounterFactory().create_anthropic_counter(),
                pricing={
                    TokenDirection.ENCODE: 15.00 / 1_000_000,
                    TokenDirection.DECODE: 75.00 / 1_000_000,
                },
            ),
            'claude-sonnet-4-0': ModelConfig(
                client_class=ChatAnthropic,
                token_counter=TokenCounterFactory().create_anthropic_counter(),
                pricing={
                    TokenDirection.ENCODE: 3.00 / 1_000_000,
                    TokenDirection.DECODE: 15.00 / 1_000_000,
                },
            ),
            # Google
            'gemini-2.0-flash-001': ModelConfig(
                client_class=ChatGoogleGenerativeAI,
                token_counter=TokenCounterFactory().create_google_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.10 / 1_000_000,
                    TokenDirection.DECODE: 0.40 / 1_000_000,
                },
            ),
            'gemini-2.5-flash': ModelConfig(
                client_class=ChatGoogleGenerativeAI,
                token_counter=TokenCounterFactory().create_google_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.30 / 1_000_000,
                    TokenDirection.DECODE: 1.00 / 1_000_000,
                },
            ),
            'gemini-2.5-pro-preview-06-05': ModelConfig(
                client_class=ChatGoogleGenerativeAI,
                token_counter=TokenCounterFactory().create_google_counter(),
                pricing={
                    TokenDirection.ENCODE: 2.50 / 1_000_000,
                    TokenDirection.DECODE: 15.00 / 1_000_000,
                },
            ),
            # Groq
            'grok-3-mini': ModelConfig(
                client_class=ChatXAI,
                token_counter=TokenCounterFactory().create_xai_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.30 / 1_000_000,
                    TokenDirection.DECODE: 0.50 / 1_000_000,
                },
            ),
            'grok-3': ModelConfig(
                client_class=ChatXAI,
                token_counter=TokenCounterFactory().create_xai_counter(),
                pricing={
                    TokenDirection.ENCODE: 3.00 / 1_000_000,
                    TokenDirection.DECODE: 15.00 / 1_000_000,
                },
            ),
            'grok-3-fast': ModelConfig(
                client_class=ChatXAI,
                token_counter=TokenCounterFactory().create_xai_counter(),
                pricing={
                    TokenDirection.ENCODE: 5.00 / 1_000_000,
                    TokenDirection.DECODE: 25.00 / 1_000_000,
                },
            ),
            # DeepSeek
            'deepseek-chat': ModelConfig(
                client_class=ChatDeepSeek,
                token_counter=TokenCounterFactory().create_deepseek_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.27 / 1_000_000,
                    TokenDirection.DECODE: 1.10 / 1_000_000,
                },
            ),
            'deepseek-reasoner': ModelConfig(
                client_class=ChatDeepSeek,
                token_counter=TokenCounterFactory().create_deepseek_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.55 / 1_000_000,
                    TokenDirection.DECODE: 2.19 / 1_000_000,
                },
            ),
            # Cerebas
            'gpt-oss-120b': ModelConfig(
                client_class=ChatCerebras,
                token_counter=TokenCounterFactory().create_cerebas_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.25 / 1_000_000,
                    TokenDirection.DECODE: 0.69 / 1_000_000,
                },
            ),
            'qwen-3-32b': ModelConfig(
                client_class=ChatCerebras,
                token_counter=TokenCounterFactory().create_cerebas_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.40 / 1_000_000,
                    TokenDirection.DECODE: 0.80 / 1_000_000,
                },
            ),
            'llama-4-scout-17b-16e-instruct': ModelConfig(
                client_class=ChatCerebras,
                token_counter=TokenCounterFactory().create_cerebas_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.65 / 1_000_000,
                    TokenDirection.DECODE: 0.85 / 1_000_000,
                },
            ),
            'llama-4-maverick-17b-128e-instruct': ModelConfig(
                client_class=ChatCerebras,
                token_counter=TokenCounterFactory().create_cerebas_counter(),
                pricing={
                    TokenDirection.ENCODE: 0.20 / 1_000_000,
                    TokenDirection.DECODE: 0.60 / 1_000_000,
                },
            ),
        }

    async def get_tokens(self, model_name: str, messages: list[BaseMessage]) -> int:
        """Получает нужную функцию счетчика токенов и вызывает ее

        Args:
            model_name (str): Название модели
            messages (list[BaseMessage]): Сообщения

        Returns:
            int: Количество токенов
        """
        if model_name not in self._models:
            raise ValueError(f'Unknown model: {model_name}')
        return await self._models[model_name].token_counter(
            messages,
            model_name,
            self.client,
        )

    async def get_moderation(
        self,
        model_name: str,
        messages: list[BaseMessage],
    ) -> None:
        """Получает нужную функцию модерации и вызывает ее

        Args:
            model_name (str): Название модели
            messages (list[BaseMessage]): Сообщения
        """
        if model_name not in self._models:
            raise ValueError(f'Unknown model: {model_name}')
        if self._models[model_name].moderation is None:
            warnings.warn('No moderation for this model')
            return None
        return await self._models[model_name].moderation(messages, self.client)

    def init_client(self, config: dict[str, Any]) -> LLMClientInstance:
        """Инициализирует клиента LLM

        Args:
            config (dict): Конфигурация

        Returns:
            LLMClientInstance: Клиент LLM
        """
        model_name = config.get('model')
        if model_name not in self._models:
            raise ValueError(f'Unknown model: {model_name}')
        self.client = self._models[model_name].client_class(**config)
        return self.client

    def get_price(self, model_name: str, direction: TokenDirection) -> float:
        """Получает нужную цену

        Args:
            model_name (str): Название модели
            direction (TokenDirection): Направление

        Returns:
            float: Цена
        """
        if model_name not in self._models:
            raise ValueError(f'Unknown model: {model_name}')
        return self._models[model_name].pricing[direction]



import asyncio

from example.common_imports import *  # noqa: F403
from llm.service import LLMService


async def chatgpt_mgr() -> None:
    llm = await LLMService.create(gpt_4o_mini.to_dict())  # noqa: F405
    async with llm.astream_mgr(message='Расскажи теорему Пифагора') as stream:
        async for chunk in stream:
            print(chunk, end='', flush=True)


async def gigachat_mgr() -> None:
    llm = await LLMService.create(giga_chat.to_dict())  # noqa: F405
    async with llm.astream_mgr(message='Расскажи теорему Пифагора') as stream:
        async for chunk in stream:
            print(chunk, end='', flush=True)


async def claude_mgr() -> None:
    llm = await LLMService.create(claude_3_5_haiku.to_dict())  # noqa: F405
    async with llm.astream_mgr(message='Расскажи теорему Пифагора') as stream:
        async for chunk in stream:
            print(chunk, end='', flush=True)


async def gemini_mgr() -> None:
    llm = await LLMService.create(gemini_2_5_pro.to_dict())  # noqa: F405
    async with llm.astream_mgr(message='Расскажи теорему Пифагора') as stream:
        async for chunk in stream:
            print(chunk, end='', flush=True)


async def grok_mgr() -> None:
    llm = await LLMService.create(grok_3_mini.to_dict())  # noqa: F405
    async with llm.astream_mgr(message='Расскажи теорему Пифагора') as stream:
        async for chunk in stream:
            print(chunk, end='', flush=True)


async def deepseek_mgr() -> None:
    llm = await LLMService.create(deepseek_chat.to_dict())  # noqa: F405
    async with llm.astream_mgr(message='Расскажи теорему Пифагора') as stream:
        async for chunk in stream:
            print(chunk, end='', flush=True)


async def cerebras_mgr() -> None:
    llm = await LLMService.create(llama_4_maverick_17b_128e_instruct.to_dict())  # noqa: F405
    async with llm.astream_mgr(message='Расскажи теорему Пифагора') as stream:
        async for chunk in stream:
            print(chunk, end='', flush=True)


async def main() -> None:
    # await chatgpt_mgr()
    # await gigachat_mgr()
    # await claude_mgr()
    # await gemini_mgr()
    # await grok_mgr()
    # await deepseek_mgr()
    await cerebras_mgr()


if __name__ == '__main__':
    asyncio.run(main())



import json
from typing import Any, Awaitable, Callable, TypeVar

from langchain.schema import AIMessage
from pydantic import BaseModel

from llm.counter import TokenCounter
from llm.pydantic_utils.checker import is_pydantic_instance

PydanticSchema = TypeVar('PydanticSchema', bound=BaseModel)


class BillingDecorator:
    """Декоратор для расчета токенов и расходов в USD при вызове LLM-функции."""

    def __init__(
        self, func: Callable[..., Awaitable[Any]], counter: TokenCounter
    ) -> None:
        self.func = func
        self.counter = counter

    def __get__(self, instance, owner):
        return lambda *args, **kwargs: self(instance, *args, **kwargs)

    async def __call__(self, *args, **kwargs) -> Any:
        result: AIMessage | PydanticSchema = await self.func(*args, **kwargs)

        if isinstance(result, AIMessage):
            await self.counter.count_input_tokens(
                result.usage_metadata['input_tokens'],
            )
            await self.counter.count_output_tokens(
                result.usage_metadata['output_tokens'],
            )
        elif is_pydantic_instance(result):
            await self.counter.count_input_tokens_from_text(
                kwargs.get('chat_for_model'),
            )

            result_dict = result.model_dump()
            output_text = json.dumps(result_dict, ensure_ascii=False)
            await self.counter.count_output_tokens_from_text(
                [AIMessage(content=output_text)],
            )

        return result



from tokenizers import Tokenizer


class DeepSeekTokenizer:
    def __init__(self, tokenizer_path='llm/tokenizer/deepseek/tokenizer.json'):
        self.tokenizer = Tokenizer.from_file(tokenizer_path)

    def count_tokens(self, text):
        return len(self.tokenizer.encode(text).ids)

    def encode(self, text):
        result = self.tokenizer.encode(text)
        return {'ids': result.ids, 'tokens': result.tokens, 'count': len(result.ids)}


# Использование
# tokenizer = DeepSeekTokenizer()
# print(tokenizer.count_tokens("Hello!"))
# print(tokenizer.encode("Hello!"))



import asyncio

from example.common_imports import *  # noqa: F403
from llm.service import LLMService


async def chatgpt() -> None:
    llm = await LLMService.create(gpt_4o_mini.to_dict())  # noqa: F405
    result = llm.astream(message='Расскажи теорему Пифагора')
    async for chunk in result:
        print(chunk, end='', flush=True)


async def gigachat() -> None:
    llm = await LLMService.create(giga_chat.to_dict())  # noqa: F405
    result = llm.astream(message='Расскажи теорему Пифагора')
    async for chunk in result:
        print(chunk, end='', flush=True)


async def claude() -> None:
    llm = await LLMService.create(claude_3_5_haiku.to_dict())  # noqa: F405
    result = llm.astream(message='Расскажи теорему Пифагора')
    async for chunk in result:
        print(chunk, end='', flush=True)


async def gemini() -> None:
    llm = await LLMService.create(gemini_2_5_pro.to_dict())  # noqa: F405
    result = llm.astream(message='Расскажи теорему Пифагора')
    async for chunk in result:
        print(chunk, end='', flush=True)


async def grok() -> None:
    llm = await LLMService.create(grok_3_mini.to_dict())  # noqa: F405
    result = llm.astream(message='Расскажи теорему Пифагора')
    async for chunk in result:
        print(chunk, end='', flush=True)


async def deepseek() -> None:
    llm = await LLMService.create(deepseek_chat.to_dict())  # noqa: F405
    result = llm.astream(message='Расскажи теорему Пифагора')
    async for chunk in result:
        print(chunk, end='', flush=True)


async def cerebras() -> None:
    llm = await LLMService.create(llama_4_maverick_17b_128e_instruct.to_dict())  # noqa: F405
    result = llm.astream(message='Расскажи теорему Пифагора')
    async for chunk in result:
        print(chunk, end='', flush=True)


async def main() -> None:
    # await chatgpt()
    # await gigachat()
    # await claude()
    # await gemini()
    # await grok()
    # await deepseek()
    await cerebras()


if __name__ == '__main__':
    asyncio.run(main())



from langchain.schema import BaseMessage
from langchain_openai import ChatOpenAI
from openai import AsyncOpenAI

LLMClientInstance = ChatOpenAI


class ModerationError(Exception):
    pass


class ModerationPrompt:
    @staticmethod
    async def openai(
        messages: list[BaseMessage],
        client: LLMClientInstance | None = None,
    ) -> None:
        dict_messages = [message.model_dump().get('content') for message in messages]

        openai_api_key = client.openai_api_key._secret_value

        local_client = AsyncOpenAI(api_key=openai_api_key)

        response = await local_client.moderations.create(
            model='omni-moderation-latest',
            input=dict_messages,
        )

        if response.results[0].flagged:
            raise ModerationError



from dataclasses import dataclass


@dataclass
class TokenUsage:
    """Класс для хранения информации об использовании токенов и стоимости."""

    all_input_tokens: int = 0
    all_output_tokens: int = 0

    last_input_tokens: int = 0
    last_output_tokens: int = 0

    all_input_spendings: float = 0
    all_output_spendings: float = 0

    last_input_spendings: float = 0
    last_output_spendings: float = 0

    @property
    def total_tokens(self):
        return self.all_input_tokens + self.all_output_tokens

    @property
    def total_spendings(self):
        return self.all_input_spendings + self.all_output_spendings

    @property
    def last_tokens(self):
        return self.last_input_tokens + self.last_output_tokens

    @property
    def last_spendings(self):
        return self.last_input_spendings + self.last_output_spendings



import asyncio

from pydantic import BaseModel, Field

from example.common_imports import *  # noqa: F403
from llm.service import LLMService


class RelatedConceptOutput(BaseModel):
    """Новый термин и его сила связи с исходным термином."""

    title: str = Field(..., description='Название термина')
    length: int = Field(..., description='Сила связи')


class RelatedConceptListOutput(BaseModel):
    """Новые термины и их сила связи с исходным термином."""

    concepts: list[RelatedConceptOutput]


SYSTEM_PROMPT = (
    'Тебе дано понятие школьной программы: "Молекула". Сгенерируй ровно "5" понятий '
    'школьной программы, наиболее близких к этому понятию.'
)


async def chatgpt() -> None:
    llm = await LLMService.create(gpt_4o_mini.to_dict())  # noqa: F405
    structured_llm = await llm.with_structured_output(RelatedConceptListOutput)
    result = await structured_llm.ainvoke(message=SYSTEM_PROMPT)
    print(result)
    print(structured_llm.usage)
    print(structured_llm.counter.model_registry.usd_rate)


async def gigachat() -> None:
    llm = await LLMService.create(giga_chat_2_max.to_dict())  # noqa: F405
    structured_llm = await llm.with_structured_output(RelatedConceptListOutput)
    result = await structured_llm.ainvoke(message=SYSTEM_PROMPT)
    print(result)
    print(structured_llm.usage)
    print(structured_llm.counter.model_registry.usd_rate)


async def claude() -> None:
    llm = await LLMService.create(claude_3_5_haiku.to_dict())  # noqa: F405
    structured_llm = await llm.with_structured_output(RelatedConceptListOutput)
    result = await structured_llm.ainvoke(message=SYSTEM_PROMPT)
    print(result)
    print(structured_llm.usage)
    print(structured_llm.counter.model_registry.usd_rate)


async def gemini() -> None:
    llm = await LLMService.create(gemini_2_5_flash.to_dict())  # noqa: F405
    structured_llm = await llm.with_structured_output(RelatedConceptListOutput)
    result = await structured_llm.ainvoke(message=SYSTEM_PROMPT)
    print(result)
    print(structured_llm.usage)
    print(structured_llm.counter.model_registry.usd_rate)


async def grok() -> None:
    llm = await LLMService.create(grok_3_mini.to_dict())  # noqa: F405
    structured_llm = await llm.with_structured_output(RelatedConceptListOutput)
    result = await structured_llm.ainvoke(message=SYSTEM_PROMPT)
    print(result)
    print(structured_llm.usage)
    print(structured_llm.counter.model_registry.usd_rate)


async def deepseek() -> None:
    llm = await LLMService.create(deepseek_chat.to_dict())  # noqa: F405
    structured_llm = await llm.with_structured_output(RelatedConceptListOutput)
    result = await structured_llm.ainvoke(message=SYSTEM_PROMPT)
    print(result)
    print(structured_llm.usage)
    print(structured_llm.counter.model_registry.usd_rate)


async def cerebras() -> None:
    llm = await LLMService.create(llama_4_maverick_17b_128e_instruct.to_dict())  # noqa: F405
    structured_llm = await llm.with_structured_output(RelatedConceptListOutput)
    result = await structured_llm.ainvoke(message=SYSTEM_PROMPT)
    print(result)
    print(structured_llm.usage)
    print(structured_llm.counter.model_registry.usd_rate)


async def main() -> None:
    # await chatgpt()
    # await gigachat()
    # await claude()
    # await gemini()
    # await grok()
    # await deepseek()
    await cerebras()


if __name__ == '__main__':
    asyncio.run(main())



from contextlib import asynccontextmanager
from functools import partial
from typing import Any, AsyncGenerator, Literal, Self

from langchain.schema import BaseMessage

from llm.billing import BillingDecorator
from llm.cbr.cbr import CBRRate
from llm.counter import TokenCounter
from llm.model_registry import ModelRegistry
from llm.prepare_chat import PrepareChat
from llm.usage import TokenUsage


class LLMService:
    """
    Класс для работы с LLM.

    После инициализации
        `llm = await LLMService.create(get_llm_config('gpt-4o-mini'))`

    будут доступны такие поля:
        `llm.usage.all_input_tokens` - общее количество отправленных токенов
            с момента инициализации
        `llm.usage.all_output_tokens`  - общее количество полученных токенов
            с момента инициализации
        `llm.usage.last_input_tokens` - количество отправленных токенов
            за последний вызов
        `llm.usage.last_output_tokens` - количество полученных токенов
            за последний вызов

        ---

        `llm.usage.all_input_spendings` - общие расходы в USD при отправке
            с момента инициализации
        `llm.usage.all_output_spendings` - общие расходы в USD при получении
            с момента инициализации
        `llm.usage.last_input_spendings` - расходы в USD при отправке
            за последний вызов
        `llm.usage.last_output_spendings` - расходы в USD при получении
            за последний вызов

        ---

        `llm.counter.model_registry.usd_rate` - курс валюты в USD.
    """

    def __init__(self, config: dict, usd_rate: float = None) -> None:
        self.config = config
        self.model_registry = ModelRegistry(usd_rate)

        self.client = self.model_registry.init_client(config)
        self.usage = TokenUsage()
        self._is_structured_output = False

        # Инициализируем pricing с предоставленным курсом доллара
        self.counter = TokenCounter(
            model_name=config.get('model'),
            usage=self.usage,
            model_registry=self.model_registry,
        )

        self.__ainvoke = partial(BillingDecorator(self.__ainvoke, self.counter))

    @classmethod
    async def create(cls, config: dict) -> Self:
        # Получаем курс доллара
        cbr = CBRRate()
        usd_rate = await cbr.get_usd_rate()

        # Создаем экземпляр класса с уже полученным курсом
        instance = cls(config, usd_rate)
        return instance

    async def _moderation_check(
        self,
        moderation: bool,
        chat_for_model: list[BaseMessage],
    ) -> None:
        if moderation:
            await self.model_registry.get_moderation(
                self.config.get('model'),
                chat_for_model,
            )

    async def ainvoke(
        self,
        chat_history: list[dict[str, str] | BaseMessage] | None = None,
        system_prompt: str | BaseMessage | None = None,
        message: str | BaseMessage | None = None,
        moderation: bool = True,
        **kwargs,
    ) -> str:
        chat_for_model = PrepareChat(chat_history, system_prompt, message)

        await self._moderation_check(moderation, chat_for_model)

        result = await self.__ainvoke(chat_for_model=chat_for_model, **kwargs)

        if self._is_structured_output:
            return result

        return result.content

    async def __ainvoke(
        self, *, chat_for_model: list[BaseMessage], **kwargs
    ) -> BaseMessage:
        return await self.client.ainvoke(chat_for_model, **kwargs)

    async def with_structured_output(
        self,
        schema: dict | type,
        *,
        method: Literal[
            'function_calling', 'json_mode', 'format_instructions'
        ] = 'function_calling',
        include_raw: bool = False,
        **kwargs: Any,
    ) -> Self:
        new_instance = await LLMService.create(self.config)

        new_instance.client = self.client.with_structured_output(
            schema, method=method, include_raw=include_raw, **kwargs
        )
        new_instance._is_structured_output = True

        return new_instance

    async def astream(
        self,
        chat_history: list[dict[str, str] | BaseMessage] | None = None,
        system_prompt: str | BaseMessage | None = None,
        message: str | BaseMessage | None = None,
        moderation: bool = True,
        **kwargs,
    ) -> AsyncGenerator[str, None]:
        chat_for_model = PrepareChat(chat_history, system_prompt, message)

        await self._moderation_check(moderation, chat_for_model)

        async for chunk in self.__astream(chat_for_model=chat_for_model, **kwargs):
            if hasattr(chunk, 'content') and chunk.content:
                yield chunk.content

    async def __astream(self, *, chat_for_model: list[BaseMessage], **kwargs):
        async for chunk in self.client.astream(chat_for_model, **kwargs):
            yield chunk

    @asynccontextmanager
    async def astream_mgr(
        self,
        chat_history: list[dict[str, str] | BaseMessage] | None = None,
        system_prompt: str | BaseMessage | None = None,
        message: str | BaseMessage | None = None,
        **kwargs,
    ) -> AsyncGenerator[AsyncGenerator[str, None], None]:
        """
        Контекстный менеджер для потоковой генерации описания.

        Пример использования:
        ```
        async with llm.astream_mgr(system_prompt=prompt) as stream:
            async for chunk in stream:
                # обработка chunk
        ```
        """
        try:
            generator = self.astream(
                chat_history=chat_history,
                system_prompt=system_prompt,
                message=message,
                **kwargs,
            )

            yield generator

        finally:
            pass
